package core.schedules;

import core.api.models.timemanager.taskSchedule.scheduleInfos.AdvancedScheduleFieldInfo;
import core.api.models.timemanager.taskSchedule.scheduleInfos.AdvancedScheduleInfo;
import core.api.models.timemanager.taskSchedule.scheduleInfos.BasicScheduleFieldInfo;
import core.api.models.timemanager.taskSchedule.ForgottenScheduleInfo;
import core.entities.timemanager.AbstractSchedule;
import core.entities.timemanager.AdvancedTaskSchedule;
import core.entities.timemanager.BasicTaskSchedule;
import core.entities.timemanager.Task;
import core.services.ServiceExitCode;
import core.services.ServiceResult;
import core.services.TaskScheduleService;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.test.context.jdbc.Sql;
import org.springframework.test.context.jdbc.SqlGroup;

import javax.persistence.EntityManager;
import javax.transaction.Transactional;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.List;

import static org.assertj.core.api.AssertionsForClassTypes.assertThat;
import static org.junit.jupiter.api.Assertions.*;

@SpringBootTest
@Transactional
public class ScheduleServiceTest {


    @Autowired
    private EntityManager entityManager;

    @Autowired private TaskScheduleService taskScheduleService;

    private static final String username = "Testuser1";
    private static final String username2 = "Testuser2";

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void getAllSchedulesOfUser() {
        assertEquals(0, taskScheduleService.getAllSchedulesOfUser(username2).size());

        List<AbstractSchedule> result_1 = taskScheduleService.getAllSchedulesOfUser(username);
        assertEquals(5, result_1.size());
        assertTrue(result_1.contains(entityManager.find(AbstractSchedule.class, 1L)));
        assertTrue(result_1.contains(entityManager.find(AbstractSchedule.class, 2L)));
        assertTrue(result_1.contains(entityManager.find(AbstractSchedule.class, 3L)));
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void scheduleBasic() {
        //Situation 1: Schedule finished Task
        ServiceResult<AbstractSchedule> result_1 = taskScheduleService.scheduleBasic(entityManager.find(Task.class, 2L), new BasicScheduleFieldInfo(LocalDate.now()));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_1.getExitCode());

        //Situation 2: Schedule before today
        ServiceResult<AbstractSchedule> result_2 = taskScheduleService.scheduleBasic(entityManager.find(Task.class, 1L), new BasicScheduleFieldInfo(LocalDate.of(2010, 3, 14)));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_2.getExitCode());

        //Situation 3: Valid schedule
        ServiceResult<AbstractSchedule> result_3 = taskScheduleService.scheduleBasic(entityManager.find(Task.class, 1L), new BasicScheduleFieldInfo(LocalDate.now()));
        assertEquals(ServiceExitCode.OK, result_3.getExitCode());
        assertThat(entityManager.find(BasicTaskSchedule.class, result_3.getResult().getScheduleID())).isNotNull();
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void editBasicSchedule() {
        //Situation 1: Reschedule finished task
        ServiceResult<AbstractSchedule> result_1 = taskScheduleService.editBasicSchedule(entityManager.find(BasicTaskSchedule.class, 2L), new BasicScheduleFieldInfo(LocalDate.now()));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_1.getExitCode());

        //Situation 2: Reschedule unfinished task with invalid reschedule date
        ServiceResult<AbstractSchedule> result_2 = taskScheduleService.editBasicSchedule(entityManager.find(BasicTaskSchedule.class, 1L), new BasicScheduleFieldInfo(LocalDate.of(2011, 3, 4)));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_2.getExitCode());

        //Situation 3: Reschedule unfinished task with valid reschedule date
        LocalDate oldDate = entityManager.find(BasicTaskSchedule.class, 1L).getScheduleDate();
        ServiceResult<AbstractSchedule> result_3 = taskScheduleService.editBasicSchedule(entityManager.find(BasicTaskSchedule.class, 1L), new BasicScheduleFieldInfo(LocalDate.now()));
        assertEquals(ServiceExitCode.OK, result_3.getExitCode());
        assertNotEquals(((BasicTaskSchedule) result_3.getResult()).getScheduleDate(), oldDate);

        //Situation 4: Reschedule already running schedule
        ServiceResult<AbstractSchedule> result_4 = taskScheduleService.editBasicSchedule(entityManager.find(BasicTaskSchedule.class, 4L),
                new BasicScheduleFieldInfo(LocalDate.now()));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_4.getExitCode());
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void deleteSchedule() {
        for(long i=1; i<=2; i++) {
            taskScheduleService.deleteSchedule(entityManager.find(BasicTaskSchedule.class, i));
            assertThat(entityManager.find(BasicTaskSchedule.class, i)).isNull();
        }
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void getFilteredScheduledOfUser() {
        //Invalid user
        assertEquals(0, taskScheduleService.getFilteredScheduledOfUser(LocalDate.now(), false, "Quatsch").size());
        assertEquals(0, taskScheduleService.getFilteredScheduledOfUser(LocalDate.now(), true, "Quatsch").size());

        //User with no tasks/schedules
        assertEquals(0, taskScheduleService.getFilteredScheduledOfUser(LocalDate.now(), false, username2).size());
        assertEquals(0, taskScheduleService.getFilteredScheduledOfUser(LocalDate.now(), true, username2).size());

        //user with tasks and schedules
        List<AbstractSchedule> result_1 = taskScheduleService.getFilteredScheduledOfUser(LocalDate.of(2024,11,11), false, username);
        assertEquals(3, result_1.size());
        assertTrue(result_1.contains(entityManager.find(BasicTaskSchedule.class, 3L)));
        assertTrue(result_1.contains(entityManager.find(BasicTaskSchedule.class, 4L)));
        assertTrue(result_1.contains(entityManager.find(BasicTaskSchedule.class, 5L)));

        List<AbstractSchedule> result_2 = taskScheduleService.getFilteredScheduledOfUser(LocalDate.of(2024,11,11), true, username);
        assertEquals(1, result_2.size());
    }
    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void scheduleNow() {
        //Situation 1: Task has already an active schedule
        ServiceResult<AbstractSchedule> result_1 = taskScheduleService.scheduleNow(entityManager.find(Task.class, 3L));
        assertEquals(ServiceExitCode.ENTITY_ALREADY_EXIST, result_1.getExitCode());

        //Situation 2: Task is already finished
        ServiceResult<AbstractSchedule> result_2 = taskScheduleService.scheduleNow(entityManager.find(Task.class, 2L));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_2.getExitCode());

        //Situation 3: Task can be scheduled and started
        ServiceResult<AbstractSchedule> result_3 = taskScheduleService.scheduleNow(entityManager.find(Task.class, 5L));
        assertEquals(ServiceExitCode.OK, result_3.getExitCode());
        assertThat(entityManager.find(BasicTaskSchedule.class, result_3.getResult().getScheduleID())).isNotNull();

        //Situation 4: Running Advanced Schedule
        ServiceResult<AbstractSchedule> result_4 = taskScheduleService.scheduleNow(entityManager.find(Task.class, 17L));
        assertEquals(ServiceExitCode.ENTITY_ALREADY_EXIST, result_4.getExitCode());
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void activateSchedule() {
        //Activate already running schedule
        ServiceResult<AbstractSchedule> result_1 = taskScheduleService.activateSchedule(entityManager.find(BasicTaskSchedule.class, 4L));
        assertNotNull(result_1.getResult().getStartTime());
        assertNull(result_1.getResult().getStopTime());

        entityManager.remove(entityManager.find(BasicTaskSchedule.class, 4L));
        assertThat(entityManager.find(BasicTaskSchedule.class, 4L)).isNull();

        //Activate not running schedule
        ServiceResult<AbstractSchedule> result_2 = taskScheduleService.activateSchedule(entityManager.find(BasicTaskSchedule.class, 5L));
        assertNotNull(result_2.getResult().getStartTime());
        assertNull(result_2.getResult().getStopTime());
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void stopSchedule() {
        //Stop schedule that is not running 4
        ServiceResult<AbstractSchedule> result_1 = taskScheduleService.stopSchedule(entityManager.find(BasicTaskSchedule.class, 5L), false);
        assertNull(result_1.getResult().getStartTime());
        assertNull(result_1.getResult().getStopTime());

        //Stop schedule (without finishing)
        ServiceResult<AbstractSchedule> result_2 = taskScheduleService.stopSchedule(entityManager.find(BasicTaskSchedule.class, 4L), false);
        assertNotNull(result_2.getResult().getStartTime());
        assertNotNull(result_2.getResult().getStopTime());

        //Stop schedule with finishing
        ServiceResult<AbstractSchedule> result_3 = taskScheduleService.stopSchedule(entityManager.find(BasicTaskSchedule.class, 7L), true);
        assertNotNull(result_3.getResult().getStartTime());
        assertNotNull(result_3.getResult().getStopTime());
        assertTrue(result_3.getResult().getTask().isFinished());
        assertFalse(result_3.getResult().isStartable());
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void registerForgottenSchedule() {
        //Register task schedule for task that is already finished
        LocalDateTime startTime = LocalDateTime.now().minusMinutes(10L);
        LocalDateTime finishTime = LocalDateTime.now();

        ServiceResult<AbstractSchedule> result_1 = taskScheduleService.registerForgottenSchedule(entityManager.find(Task.class, 2L), new ForgottenScheduleInfo(startTime, finishTime));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_1.getExitCode());

        ServiceResult<AbstractSchedule> result_2 = taskScheduleService.registerForgottenSchedule(entityManager.find(Task.class, 5L), new ForgottenScheduleInfo(startTime, finishTime));
        assertEquals(ServiceExitCode.OK, result_2.getExitCode());
        assertEquals(startTime, result_2.getResult().getStartTime());
        assertEquals(finishTime, result_2.getResult().getStopTime());
        assertThat(entityManager.find(BasicTaskSchedule.class, result_2.getResult().getScheduleID())).isNotNull();
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void getAllMissedSchedulesOfUser() {
        assertEquals(1, taskScheduleService.getAllMissedSchedulesOfUser(username).size());
    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void scheduleAdvanced() {
        //Situation 1: Schedule finished Task
        ServiceResult<AbstractSchedule> result_1 = taskScheduleService.scheduleAdvanced(entityManager.find(Task.class, 18L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now().plusHours(1L), LocalDateTime.now().plusHours(2L)));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_1.getExitCode());

        //Situation 2: Schedule Start is before today
        ServiceResult<AbstractSchedule> result_2 = taskScheduleService.scheduleAdvanced(entityManager.find(Task.class, 17L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now().minusDays(1L), LocalDateTime.now().plusHours(2L)));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_2.getExitCode());

        //Situation 3: Schedule End is before today
        ServiceResult<AbstractSchedule> result_3 = taskScheduleService.scheduleAdvanced(entityManager.find(Task.class, 17L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now().minusDays(2L), LocalDateTime.now().minusDays(1L)));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_3.getExitCode());

        //Situation 4: Start after stop
        ServiceResult<AbstractSchedule> result_4 = taskScheduleService.scheduleAdvanced(entityManager.find(Task.class, 17L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now().plusHours(2L), LocalDateTime.now().plusHours(1L)));
        assertEquals(ServiceExitCode.INVALID_PARAMETER, result_4.getExitCode());

        //Situation 5: Valid schedule
        ServiceResult<AbstractSchedule> result_5 = taskScheduleService.scheduleAdvanced(entityManager.find(Task.class, 17L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now(), LocalDateTime.now().plusHours(1L)));
        assertEquals(ServiceExitCode.OK, result_5.getExitCode());
        assertThat(entityManager.find(AdvancedTaskSchedule.class, result_5.getResult().getScheduleID())).isNotNull();


    }

    @Test
    @SqlGroup({
            @Sql("classpath:taskgroupRepositoryTestEntries.sql"),
            @Sql("classpath:taskRepositoryEntries.sql"),
            @Sql("classpath:basicScheduleEntries.sql")
    })
    void editscheduleAdvanced() {
        //Situation 1: Schedule finished Task
        ServiceResult<AbstractSchedule> result_1 = taskScheduleService.editAdvancedSchedule(entityManager.find(AdvancedTaskSchedule.class, 12L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now().plusHours(1L), LocalDateTime.now().plusHours(2L)));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_1.getExitCode());

        //Situation 2: Schedule Start is before today
        ServiceResult<AbstractSchedule> result_2 = taskScheduleService.editAdvancedSchedule(entityManager.find(AdvancedTaskSchedule.class, 11L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now().minusDays(1L), LocalDateTime.now().plusHours(2L)));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_2.getExitCode());

        //Situation 3: Schedule End is before today
        ServiceResult<AbstractSchedule> result_3 = taskScheduleService.editAdvancedSchedule(entityManager.find(AdvancedTaskSchedule.class, 11L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now().minusDays(2L), LocalDateTime.now().minusDays(1L)));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_3.getExitCode());

        //Situation 4: Start after stop
        ServiceResult<AbstractSchedule> result_4 = taskScheduleService.editAdvancedSchedule(entityManager.find(AdvancedTaskSchedule.class, 11L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now().plusHours(2L), LocalDateTime.now().plusHours(1L)));
        assertEquals(ServiceExitCode.INVALID_PARAMETER, result_4.getExitCode());

        //Situation 5: Valid schedule
        ServiceResult<AbstractSchedule> result_5 = taskScheduleService.editAdvancedSchedule(entityManager.find(AdvancedTaskSchedule.class, 11L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now(), LocalDateTime.now().plusHours(1L)));
        assertEquals(ServiceExitCode.OK, result_5.getExitCode());
        assertThat(entityManager.find(AdvancedTaskSchedule.class, result_5.getResult().getScheduleID())).isNotNull();

        //Situation 6: reschedule already running schedule
        ServiceResult<AbstractSchedule> result_6 = taskScheduleService.editAdvancedSchedule(entityManager.find(AdvancedTaskSchedule.class, 9L),
                new AdvancedScheduleFieldInfo(LocalDateTime.now(), LocalDateTime.now().plusHours(1L)));
        assertEquals(ServiceExitCode.INVALID_OPERATION, result_6.getExitCode());
    }
}
